// SPDX-License-Identifier: GPL-3.0-or-later
// Copyright (c) 2025 Abel Luck <abel@guardianproject.info>

package provider

import (
	"encoding/base64"
	"encoding/hex"
	"encoding/json"
	"fmt"
	"regexp"
	"testing"

	"github.com/hashicorp/terraform-plugin-framework/types"
	"github.com/hashicorp/terraform-plugin-testing/helper/resource"
	"github.com/hashicorp/terraform-plugin-testing/terraform"
)

func TestTorObfs4StateResource(t *testing.T) {
	resource.Test(t, resource.TestCase{
		PreCheck:                 func() { testAccPreCheck(t) },
		ProtoV6ProviderFactories: testAccProtoV6ProviderFactories,
		Steps: []resource.TestStep{
			// Create and Read testing
			{
				Config: testTorObfs4StateResourceConfigWithIdentityKeys(),
				Check: resource.ComposeAggregateTestCheckFunc(
					// Check that the resource was created
					resource.TestCheckResourceAttrSet("tor_obfs4_state.test", "id"),

					// Check that all fields are generated
					resource.TestCheckResourceAttrSet("tor_obfs4_state.test", "node_id"),
					resource.TestCheckResourceAttrSet("tor_obfs4_state.test", "private_key"),
					resource.TestCheckResourceAttrSet("tor_obfs4_state.test", "public_key"),
					resource.TestCheckResourceAttrSet("tor_obfs4_state.test", "drbg_seed"),
					resource.TestCheckResourceAttrSet("tor_obfs4_state.test", "certificate"),
					resource.TestCheckResourceAttrSet("tor_obfs4_state.test", "state_json"),
					resource.TestCheckResourceAttrSet("tor_obfs4_state.test", "bridge_line"),

					// Check default iat_mode
					resource.TestCheckResourceAttr("tor_obfs4_state.test", "iat_mode", "0"),

					// Check field lengths (hex-encoded)
					resource.TestMatchResourceAttr("tor_obfs4_state.test", "node_id", regexp.MustCompile(`^[0-9a-f]{40}$`)),
					resource.TestMatchResourceAttr("tor_obfs4_state.test", "private_key", regexp.MustCompile(`^[0-9a-f]{64}$`)),
					resource.TestMatchResourceAttr("tor_obfs4_state.test", "public_key", regexp.MustCompile(`^[0-9a-f]{64}$`)),
					resource.TestMatchResourceAttr("tor_obfs4_state.test", "drbg_seed", regexp.MustCompile(`^[0-9a-f]{48}$`)),

					// Check certificate format (base64 without padding)
					resource.TestMatchResourceAttr("tor_obfs4_state.test", "certificate", regexp.MustCompile(`^[A-Za-z0-9+/]+$`)),

					// Check bridge_line format
					resource.TestMatchResourceAttr("tor_obfs4_state.test", "bridge_line", regexp.MustCompile(`^Bridge obfs4 <IP:PORT> <FINGERPRINT> cert=[A-Za-z0-9+/]+ iat-mode=0$`)),

					// Check that the JSON is valid
					testTorObfs4StateValidateJSON("tor_obfs4_state.test"),

					// Check that certificate is properly generated from node_id + public_key
					testTorObfs4StateCertificateValidity("tor_obfs4_state.test"),

					// Check that bridge_line contains the correct certificate and iat_mode
					testTorObfs4StateBridgeLineValidity("tor_obfs4_state.test"),

					// Verify the external identity keys are being used
					resource.TestCheckResourceAttrPair("tor_obfs4_state.test", "rsa_identity_private_key", "tor_relay_identity_rsa.test", "private_key_pem"),
					resource.TestCheckResourceAttrPair("tor_obfs4_state.test", "ed25519_identity_private_key", "tor_relay_identity_ed25519.test", "private_key_pem"),
				),
			},
			// Update testing (iat_mode)
			{
				Config: testTorObfs4StateResourceConfigWithIATMode(1),
				Check: resource.ComposeAggregateTestCheckFunc(
					// Check that iat_mode was updated
					resource.TestCheckResourceAttr("tor_obfs4_state.test", "iat_mode", "1"),

					// Check that crypto fields remain unchanged (immutable)
					resource.TestCheckResourceAttrSet("tor_obfs4_state.test", "node_id"),
					resource.TestCheckResourceAttrSet("tor_obfs4_state.test", "private_key"),
					resource.TestCheckResourceAttrSet("tor_obfs4_state.test", "public_key"),
					resource.TestCheckResourceAttrSet("tor_obfs4_state.test", "drbg_seed"),
					resource.TestCheckResourceAttrSet("tor_obfs4_state.test", "certificate"),

					// Check that bridge_line reflects the updated iat_mode
					resource.TestMatchResourceAttr("tor_obfs4_state.test", "bridge_line", regexp.MustCompile(`^Bridge obfs4 <IP:PORT> <FINGERPRINT> cert=[A-Za-z0-9+/]+ iat-mode=1$`)),

					// Check that state_json reflects the updated iat_mode
					testTorObfs4StateValidateJSON("tor_obfs4_state.test"),

					// Check that bridge_line contains the correct certificate and iat_mode
					testTorObfs4StateBridgeLineValidity("tor_obfs4_state.test"),
				),
			},
		},
	})
}

func TestTorObfs4StateResourceWithCustomIATMode(t *testing.T) {
	resource.Test(t, resource.TestCase{
		PreCheck:                 func() { testAccPreCheck(t) },
		ProtoV6ProviderFactories: testAccProtoV6ProviderFactories,
		Steps: []resource.TestStep{
			{
				Config: testTorObfs4StateResourceConfigWithIATMode(2),
				Check: resource.ComposeAggregateTestCheckFunc(
					resource.TestCheckResourceAttr("tor_obfs4_state.test", "iat_mode", "2"),
					// Check that bridge_line reflects the custom iat_mode
					resource.TestMatchResourceAttr("tor_obfs4_state.test", "bridge_line", regexp.MustCompile(`^Bridge obfs4 <IP:PORT> <FINGERPRINT> cert=[A-Za-z0-9+/]+ iat-mode=2$`)),
					testTorObfs4StateValidateJSON("tor_obfs4_state.test"),
					testTorObfs4StateBridgeLineValidity("tor_obfs4_state.test"),
				),
			},
		},
	})
}

func TestTorObfs4StateResourceMultiple(t *testing.T) {
	resource.Test(t, resource.TestCase{
		PreCheck:                 func() { testAccPreCheck(t) },
		ProtoV6ProviderFactories: testAccProtoV6ProviderFactories,
		Steps: []resource.TestStep{
			{
				Config: testTorObfs4StateResourceConfigMultiple(),
				Check: resource.ComposeAggregateTestCheckFunc(
					// Check first resource
					resource.TestCheckResourceAttrSet("tor_obfs4_state.test1", "node_id"),
					resource.TestCheckResourceAttrSet("tor_obfs4_state.test1", "certificate"),
					resource.TestCheckResourceAttrSet("tor_obfs4_state.test1", "bridge_line"),

					// Check second resource
					resource.TestCheckResourceAttrSet("tor_obfs4_state.test2", "node_id"),
					resource.TestCheckResourceAttrSet("tor_obfs4_state.test2", "certificate"),
					resource.TestCheckResourceAttrSet("tor_obfs4_state.test2", "bridge_line"),

					// Verify they have different keys (uniqueness)
					testTorObfs4StateResourcesAreDifferent("tor_obfs4_state.test1", "tor_obfs4_state.test2"),
				),
			},
		},
	})
}

// Unit tests for helper functions
func TestTorObfs4StateGenerateCertificate(t *testing.T) {
	resource := &TorObfs4StateResource{}

	// Test with known values
	nodeIdHex := "9be04232642c50ee2864c2724500870f5fce6d2d"
	publicKeyHex := "45c9411955294f9b4306e4a65941ea546dee63e1c8a83dd984d1095c4a2f3911"

	cert, err := resource.generateCertificate(nodeIdHex, publicKeyHex)
	if err != nil {
		t.Fatalf("generateCertificate failed: %v", err)
	}

	// Certificate should be base64 encoded 52 bytes (node_id + public_key)
	if len(cert) != 70 { // 52 bytes base64 encoded without padding = 70 chars
		t.Errorf("Expected certificate length 70, got %d", len(cert))
	}

	// Verify the certificate decodes correctly
	decoded, err := base64.StdEncoding.DecodeString(cert + "==") // Add padding for decode
	if err != nil {
		t.Fatalf("Certificate is not valid base64: %v", err)
	}

	if len(decoded) != 52 {
		t.Errorf("Expected decoded certificate length 52, got %d", len(decoded))
	}

	// Verify node_id and public_key are correctly concatenated
	nodeIdBytes, _ := hex.DecodeString(nodeIdHex)
	publicKeyBytes, _ := hex.DecodeString(publicKeyHex)

	for i := 0; i < 20; i++ {
		if decoded[i] != nodeIdBytes[i] {
			t.Errorf("Node ID mismatch at byte %d", i)
		}
	}

	for i := 0; i < 32; i++ {
		if decoded[20+i] != publicKeyBytes[i] {
			t.Errorf("Public key mismatch at byte %d", i)
		}
	}
}

func TestTorObfs4StateGenerateCertificateInvalidInput(t *testing.T) {
	resource := &TorObfs4StateResource{}

	// Test with invalid node ID length
	_, err := resource.generateCertificate("invalid", "45c9411955294f9b4306e4a65941ea546dee63e1c8a83dd984d1095c4a2f3911")
	if err == nil {
		t.Error("Expected error for invalid node ID length")
	}

	// Test with invalid public key length
	_, err = resource.generateCertificate("9be04232642c50ee2864c2724500870f5fce6d2d", "invalid")
	if err == nil {
		t.Error("Expected error for invalid public key length")
	}

	// Test with non-hex input
	_, err = resource.generateCertificate("zzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzz", "45c9411955294f9b4306e4a65941ea546dee63e1c8a83dd984d1095c4a2f3911")
	if err == nil {
		t.Error("Expected error for non-hex node ID")
	}
}

func TestTorObfs4StateGenerateBridgeLine(t *testing.T) {
	resource := &TorObfs4StateResource{}

	// Test with known values
	certificate := "Y+m2Eny/3H4JeW2RwRYGlNXpdYk8MXRbRFuv5AaTZfxnAUbhEnWCTmJM+VTMssTCYrxyag"
	iatMode := int64(0)

	bridgeLine := resource.generateBridgeLine(certificate, iatMode)

	expectedBridgeLine := "Bridge obfs4 <IP:PORT> <FINGERPRINT> cert=Y+m2Eny/3H4JeW2RwRYGlNXpdYk8MXRbRFuv5AaTZfxnAUbhEnWCTmJM+VTMssTCYrxyag iat-mode=0"

	if bridgeLine != expectedBridgeLine {
		t.Errorf("Expected bridge line:\n%s\nGot:\n%s", expectedBridgeLine, bridgeLine)
	}

	// Test with different iat_mode
	iatMode = int64(2)
	bridgeLine = resource.generateBridgeLine(certificate, iatMode)
	expectedBridgeLine = "Bridge obfs4 <IP:PORT> <FINGERPRINT> cert=Y+m2Eny/3H4JeW2RwRYGlNXpdYk8MXRbRFuv5AaTZfxnAUbhEnWCTmJM+VTMssTCYrxyag iat-mode=2"

	if bridgeLine != expectedBridgeLine {
		t.Errorf("Expected bridge line with iat-mode=2:\n%s\nGot:\n%s", expectedBridgeLine, bridgeLine)
	}
}

func TestTorObfs4StateGenerateStateJson(t *testing.T) {
	resource := &TorObfs4StateResource{}

	// Create test data
	data := &TorObfs4StateResourceModel{
		NodeId:     typeStringValue("9be04232642c50ee2864c2724500870f5fce6d2d"),
		PrivateKey: typeStringValue("c858003d52fea6be52c6d7f8a3a44692fb3b9ad1bc3b7434871efd9acb8b1554"),
		PublicKey:  typeStringValue("45c9411955294f9b4306e4a65941ea546dee63e1c8a83dd984d1095c4a2f3911"),
		DrbgSeed:   typeStringValue("fae8faa79b74d08933251727aab6ca9627736a2387231b03"),
		IatMode:    typeInt64Value(0),
	}

	jsonStr, err := resource.generateStateJson(data)
	if err != nil {
		t.Fatalf("generateStateJson failed: %v", err)
	}

	// Parse the JSON to verify structure
	var parsed obfs4StateJson
	err = json.Unmarshal([]byte(jsonStr), &parsed)
	if err != nil {
		t.Fatalf("Generated JSON is invalid: %v", err)
	}

	// Verify fields
	if parsed.NodeId != "9be04232642c50ee2864c2724500870f5fce6d2d" {
		t.Errorf("Node ID mismatch in JSON")
	}
	if parsed.PrivateKey != "c858003d52fea6be52c6d7f8a3a44692fb3b9ad1bc3b7434871efd9acb8b1554" {
		t.Errorf("Private key mismatch in JSON")
	}
	if parsed.PublicKey != "45c9411955294f9b4306e4a65941ea546dee63e1c8a83dd984d1095c4a2f3911" {
		t.Errorf("Public key mismatch in JSON")
	}
	if parsed.DrbgSeed != "fae8faa79b74d08933251727aab6ca9627736a2387231b03" {
		t.Errorf("DRBG seed mismatch in JSON")
	}
	if parsed.IatMode != 0 {
		t.Errorf("IAT mode mismatch in JSON")
	}
}

// Test configuration functions
func testTorObfs4StateResourceConfigWithIdentityKeys() string {
	return `
resource "tor_relay_identity_rsa" "test" {}

resource "tor_relay_identity_ed25519" "test" {}

resource "tor_obfs4_state" "test" {
  rsa_identity_private_key     = tor_relay_identity_rsa.test.private_key_pem
  ed25519_identity_private_key = tor_relay_identity_ed25519.test.private_key_pem
}
`
}

func testTorObfs4StateResourceConfigWithIATMode(iatMode int) string {
	return fmt.Sprintf(`
resource "tor_relay_identity_rsa" "test" {}

resource "tor_relay_identity_ed25519" "test" {}

resource "tor_obfs4_state" "test" {
  rsa_identity_private_key     = tor_relay_identity_rsa.test.private_key_pem
  ed25519_identity_private_key = tor_relay_identity_ed25519.test.private_key_pem
  iat_mode = %d
}
`, iatMode)
}

func testTorObfs4StateResourceConfigMultiple() string {
	return `
resource "tor_relay_identity_rsa" "test1" {}

resource "tor_relay_identity_ed25519" "test1" {}

resource "tor_relay_identity_rsa" "test2" {}

resource "tor_relay_identity_ed25519" "test2" {}

resource "tor_obfs4_state" "test1" {
  rsa_identity_private_key     = tor_relay_identity_rsa.test1.private_key_pem
  ed25519_identity_private_key = tor_relay_identity_ed25519.test1.private_key_pem
  iat_mode = 0
}

resource "tor_obfs4_state" "test2" {
  rsa_identity_private_key     = tor_relay_identity_rsa.test2.private_key_pem
  ed25519_identity_private_key = tor_relay_identity_ed25519.test2.private_key_pem
  iat_mode = 1
}
`
}

// Custom test check functions
func testTorObfs4StateValidateJSON(resourceName string) resource.TestCheckFunc {
	return func(s *terraform.State) error {
		rs, ok := s.RootModule().Resources[resourceName]
		if !ok {
			return fmt.Errorf("Resource not found: %s", resourceName)
		}

		stateJsonStr := rs.Primary.Attributes["state_json"]
		if stateJsonStr == "" {
			return fmt.Errorf("state_json is empty")
		}

		// Parse JSON to verify it's valid
		var parsed obfs4StateJson
		err := json.Unmarshal([]byte(stateJsonStr), &parsed)
		if err != nil {
			return fmt.Errorf("state_json is not valid JSON: %v", err)
		}

		// Verify fields match attributes
		if parsed.NodeId != rs.Primary.Attributes["node_id"] {
			return fmt.Errorf("JSON node_id doesn't match attribute")
		}
		if parsed.PrivateKey != rs.Primary.Attributes["private_key"] {
			return fmt.Errorf("JSON private_key doesn't match attribute")
		}
		if parsed.PublicKey != rs.Primary.Attributes["public_key"] {
			return fmt.Errorf("JSON public_key doesn't match attribute")
		}
		if parsed.DrbgSeed != rs.Primary.Attributes["drbg_seed"] {
			return fmt.Errorf("JSON drbg_seed doesn't match attribute")
		}

		return nil
	}
}

func testTorObfs4StateCertificateValidity(resourceName string) resource.TestCheckFunc {
	return func(s *terraform.State) error {
		rs, ok := s.RootModule().Resources[resourceName]
		if !ok {
			return fmt.Errorf("Resource not found: %s", resourceName)
		}

		nodeIdHex := rs.Primary.Attributes["node_id"]
		publicKeyHex := rs.Primary.Attributes["public_key"]
		certificate := rs.Primary.Attributes["certificate"]

		// Regenerate certificate and compare
		resource := &TorObfs4StateResource{}
		expectedCert, err := resource.generateCertificate(nodeIdHex, publicKeyHex)
		if err != nil {
			return fmt.Errorf("Failed to generate expected certificate: %v", err)
		}

		if certificate != expectedCert {
			return fmt.Errorf("Certificate doesn't match expected value")
		}

		return nil
	}
}

func testTorObfs4StateBridgeLineValidity(resourceName string) resource.TestCheckFunc {
	return func(s *terraform.State) error {
		rs, ok := s.RootModule().Resources[resourceName]
		if !ok {
			return fmt.Errorf("Resource not found: %s", resourceName)
		}

		certificate := rs.Primary.Attributes["certificate"]
		iatMode := rs.Primary.Attributes["iat_mode"]
		bridgeLine := rs.Primary.Attributes["bridge_line"]

		// Expected bridge line format
		expectedBridgeLine := fmt.Sprintf("Bridge obfs4 <IP:PORT> <FINGERPRINT> cert=%s iat-mode=%s", certificate, iatMode)

		if bridgeLine != expectedBridgeLine {
			return fmt.Errorf("Bridge line doesn't match expected format.\nExpected: %s\nActual: %s", expectedBridgeLine, bridgeLine)
		}

		return nil
	}
}

func testTorObfs4StateResourcesAreDifferent(resource1, resource2 string) resource.TestCheckFunc {
	return func(s *terraform.State) error {
		rs1, ok := s.RootModule().Resources[resource1]
		if !ok {
			return fmt.Errorf("Resource not found: %s", resource1)
		}

		rs2, ok := s.RootModule().Resources[resource2]
		if !ok {
			return fmt.Errorf("Resource not found: %s", resource2)
		}

		// Check that crypto fields are different
		if rs1.Primary.Attributes["node_id"] == rs2.Primary.Attributes["node_id"] {
			return fmt.Errorf("Resources have same node_id")
		}
		if rs1.Primary.Attributes["private_key"] == rs2.Primary.Attributes["private_key"] {
			return fmt.Errorf("Resources have same private_key")
		}
		if rs1.Primary.Attributes["public_key"] == rs2.Primary.Attributes["public_key"] {
			return fmt.Errorf("Resources have same public_key")
		}
		if rs1.Primary.Attributes["certificate"] == rs2.Primary.Attributes["certificate"] {
			return fmt.Errorf("Resources have same certificate")
		}
		if rs1.Primary.Attributes["bridge_line"] == rs2.Primary.Attributes["bridge_line"] {
			return fmt.Errorf("Resources have same bridge_line")
		}

		return nil
	}
}

// Helper functions for creating types.String and types.Int64 values in tests
func typeStringValue(value string) types.String {
	return types.StringValue(value)
}

func typeInt64Value(value int64) types.Int64 {
	return types.Int64Value(value)
}
