// SPDX-License-Identifier: GPL-3.0-or-later
// Copyright (c) 2025 Abel Luck <abel@guardianproject.info>

package provider

import (
	"context"
	"crypto/ed25519"
	"crypto/rand"
	"crypto/sha256"
	"crypto/x509"
	"encoding/pem"
	"fmt"

	"github.com/hashicorp/terraform-plugin-framework/resource"
	"github.com/hashicorp/terraform-plugin-framework/resource/schema"
	"github.com/hashicorp/terraform-plugin-framework/resource/schema/planmodifier"
	"github.com/hashicorp/terraform-plugin-framework/resource/schema/stringplanmodifier"
	"github.com/hashicorp/terraform-plugin-framework/types"
	"github.com/hashicorp/terraform-plugin-log/tflog"
)

var _ resource.Resource = &TorRelayIdentityEd25519Resource{}

func NewTorRelayIdentityEd25519Resource() resource.Resource {
	return &TorRelayIdentityEd25519Resource{}
}

type TorRelayIdentityEd25519Resource struct{}

type TorRelayIdentityEd25519ResourceModel struct {
	Id                         types.String `tfsdk:"id"`
	Algorithm                  types.String `tfsdk:"algorithm"`
	PrivateKeyPem              types.String `tfsdk:"private_key_pem"`
	PublicKeyPem               types.String `tfsdk:"public_key_pem"`
	PublicKeyFingerprintSha256 types.String `tfsdk:"public_key_fingerprint_sha256"`
}

func (r *TorRelayIdentityEd25519Resource) Metadata(ctx context.Context, req resource.MetadataRequest, resp *resource.MetadataResponse) {
	resp.TypeName = req.ProviderTypeName + "_relay_identity_ed25519"
}

func (r *TorRelayIdentityEd25519Resource) Schema(ctx context.Context, req resource.SchemaRequest, resp *resource.SchemaResponse) {
	resp.Schema = schema.Schema{
		MarkdownDescription: "Generates Ed25519 private key for Tor relay identity as required by the Tor specification.",

		Attributes: map[string]schema.Attribute{
			"id": schema.StringAttribute{
				Computed:            true,
				MarkdownDescription: "Unique identifier based on public key fingerprint",
				PlanModifiers: []planmodifier.String{
					stringplanmodifier.UseStateForUnknown(),
				},
			},
			"algorithm": schema.StringAttribute{
				Computed:            true,
				MarkdownDescription: "Name of the algorithm used when generating the private key (always 'Ed25519')",
				PlanModifiers: []planmodifier.String{
					stringplanmodifier.UseStateForUnknown(),
				},
			},
			"private_key_pem": schema.StringAttribute{
				Computed:            true,
				Sensitive:           true,
				MarkdownDescription: "Private key data in PEM (RFC 1421) format",
				PlanModifiers: []planmodifier.String{
					stringplanmodifier.UseStateForUnknown(),
				},
			},
			"public_key_pem": schema.StringAttribute{
				Computed:            true,
				MarkdownDescription: "Public key data in PEM (RFC 1421) format",
				PlanModifiers: []planmodifier.String{
					stringplanmodifier.UseStateForUnknown(),
				},
			},
			"public_key_fingerprint_sha256": schema.StringAttribute{
				Computed:            true,
				MarkdownDescription: "SHA256 fingerprint of the public key in hex format",
				PlanModifiers: []planmodifier.String{
					stringplanmodifier.UseStateForUnknown(),
				},
			},
		},
	}
}

func (r *TorRelayIdentityEd25519Resource) Configure(ctx context.Context, req resource.ConfigureRequest, resp *resource.ConfigureResponse) {
}

func (r *TorRelayIdentityEd25519Resource) Create(ctx context.Context, req resource.CreateRequest, resp *resource.CreateResponse) {
	var data TorRelayIdentityEd25519ResourceModel

	resp.Diagnostics.Append(req.Plan.Get(ctx, &data)...)
	if resp.Diagnostics.HasError() {
		return
	}

	// Set algorithm
	data.Algorithm = types.StringValue("Ed25519")

	// Generate Ed25519 key pair
	publicKey, privateKey, err := ed25519.GenerateKey(rand.Reader)
	if err != nil {
		resp.Diagnostics.AddError("Ed25519 Key Generation Error",
			fmt.Sprintf("Unable to generate Ed25519 private key: %s", err))
		return
	}

	// Encode private key as PEM
	privateKeyPem, err := r.encodePrivateKeyPEM(privateKey)
	if err != nil {
		resp.Diagnostics.AddError("Private Key Encoding Error",
			fmt.Sprintf("Unable to encode private key as PEM: %s", err))
		return
	}
	data.PrivateKeyPem = types.StringValue(privateKeyPem)

	// Encode public key as PEM
	publicKeyPem, err := r.encodePublicKeyPEM(publicKey)
	if err != nil {
		resp.Diagnostics.AddError("Public Key Encoding Error",
			fmt.Sprintf("Unable to encode public key as PEM: %s", err))
		return
	}
	data.PublicKeyPem = types.StringValue(publicKeyPem)

	// Generate SHA256 fingerprint
	sha256Fingerprint := r.generateSha256Fingerprint(publicKey)
	data.PublicKeyFingerprintSha256 = types.StringValue(sha256Fingerprint)

	// Generate ID from SHA256 fingerprint
	data.Id = types.StringValue(fmt.Sprintf("ed25519-%s", sha256Fingerprint[:16]))

	tflog.Trace(ctx, "created tor relay identity Ed25519 resource")

	resp.Diagnostics.Append(resp.State.Set(ctx, &data)...)
}

func (r *TorRelayIdentityEd25519Resource) Read(ctx context.Context, req resource.ReadRequest, resp *resource.ReadResponse) {
	var data TorRelayIdentityEd25519ResourceModel

	resp.Diagnostics.Append(req.State.Get(ctx, &data)...)
	if resp.Diagnostics.HasError() {
		return
	}

	resp.Diagnostics.Append(resp.State.Set(ctx, &data)...)
}

func (r *TorRelayIdentityEd25519Resource) Update(ctx context.Context, req resource.UpdateRequest, resp *resource.UpdateResponse) {
	resp.Diagnostics.AddError("Update Not Supported",
		"Ed25519 relay identity keys are immutable. Any changes require resource replacement.")
}

func (r *TorRelayIdentityEd25519Resource) Delete(ctx context.Context, req resource.DeleteRequest, resp *resource.DeleteResponse) {
	tflog.Trace(ctx, "deleted tor relay identity Ed25519 resource")
}

func (r *TorRelayIdentityEd25519Resource) encodePrivateKeyPEM(privateKey ed25519.PrivateKey) (string, error) {
	privateKeyBytes, err := x509.MarshalPKCS8PrivateKey(privateKey)
	if err != nil {
		return "", err
	}
	privateKeyPem := pem.EncodeToMemory(&pem.Block{
		Type:  "PRIVATE KEY",
		Bytes: privateKeyBytes,
	})
	return string(privateKeyPem), nil
}

func (r *TorRelayIdentityEd25519Resource) encodePublicKeyPEM(publicKey ed25519.PublicKey) (string, error) {
	publicKeyBytes, err := x509.MarshalPKIXPublicKey(publicKey)
	if err != nil {
		return "", err
	}
	publicKeyPem := pem.EncodeToMemory(&pem.Block{
		Type:  "PUBLIC KEY",
		Bytes: publicKeyBytes,
	})
	return string(publicKeyPem), nil
}

func (r *TorRelayIdentityEd25519Resource) generateSha256Fingerprint(publicKey ed25519.PublicKey) string {
	publicKeyBytes, _ := x509.MarshalPKIXPublicKey(publicKey)
	sha256Sum := sha256.Sum256(publicKeyBytes)
	return fmt.Sprintf("%x", sha256Sum)
}
