// SPDX-License-Identifier: GPL-3.0-or-later
// Copyright (c) 2025 Abel Luck <abel@guardianproject.info>

package provider

import (
	"context"
	"crypto/rand"
	"crypto/rsa"
	"crypto/sha1"
	"crypto/sha256"
	"crypto/x509"
	"encoding/pem"
	"fmt"

	"github.com/hashicorp/terraform-plugin-framework/resource"
	"github.com/hashicorp/terraform-plugin-framework/resource/schema"
	"github.com/hashicorp/terraform-plugin-framework/resource/schema/planmodifier"
	"github.com/hashicorp/terraform-plugin-framework/resource/schema/stringplanmodifier"
	"github.com/hashicorp/terraform-plugin-framework/types"
	"github.com/hashicorp/terraform-plugin-log/tflog"
)

var _ resource.Resource = &TorRelayIdentityRsaResource{}

func NewTorRelayIdentityRsaResource() resource.Resource {
	return &TorRelayIdentityRsaResource{}
}

type TorRelayIdentityRsaResource struct{}

type TorRelayIdentityRsaResourceModel struct {
	Id                         types.String `tfsdk:"id"`
	Algorithm                  types.String `tfsdk:"algorithm"`
	PrivateKeyPem              types.String `tfsdk:"private_key_pem"`
	PublicKeyPem               types.String `tfsdk:"public_key_pem"`
	PublicKeyFingerprintSha1   types.String `tfsdk:"public_key_fingerprint_sha1"`
	PublicKeyFingerprintSha256 types.String `tfsdk:"public_key_fingerprint_sha256"`
}

func (r *TorRelayIdentityRsaResource) Metadata(ctx context.Context, req resource.MetadataRequest, resp *resource.MetadataResponse) {
	resp.TypeName = req.ProviderTypeName + "_relay_identity_rsa"
}

func (r *TorRelayIdentityRsaResource) Schema(ctx context.Context, req resource.SchemaRequest, resp *resource.SchemaResponse) {
	resp.Schema = schema.Schema{
		MarkdownDescription: "Generates 1024-bit RSA private key for Tor relay identity as required by the Tor specification.",

		Attributes: map[string]schema.Attribute{
			"id": schema.StringAttribute{
				Computed:            true,
				MarkdownDescription: "Unique identifier based on public key fingerprint",
				PlanModifiers: []planmodifier.String{
					stringplanmodifier.UseStateForUnknown(),
				},
			},
			"algorithm": schema.StringAttribute{
				Computed:            true,
				MarkdownDescription: "Name of the algorithm used when generating the private key (always 'RSA')",
				PlanModifiers: []planmodifier.String{
					stringplanmodifier.UseStateForUnknown(),
				},
			},
			"private_key_pem": schema.StringAttribute{
				Computed:            true,
				Sensitive:           true,
				MarkdownDescription: "Private key data in PEM (RFC 1421) format",
				PlanModifiers: []planmodifier.String{
					stringplanmodifier.UseStateForUnknown(),
				},
			},
			"public_key_pem": schema.StringAttribute{
				Computed:            true,
				MarkdownDescription: "Public key data in PEM (RFC 1421) format",
				PlanModifiers: []planmodifier.String{
					stringplanmodifier.UseStateForUnknown(),
				},
			},
			"public_key_fingerprint_sha1": schema.StringAttribute{
				Computed:            true,
				MarkdownDescription: "SHA1 fingerprint of the public key in hex format",
				PlanModifiers: []planmodifier.String{
					stringplanmodifier.UseStateForUnknown(),
				},
			},
			"public_key_fingerprint_sha256": schema.StringAttribute{
				Computed:            true,
				MarkdownDescription: "SHA256 fingerprint of the public key in hex format",
				PlanModifiers: []planmodifier.String{
					stringplanmodifier.UseStateForUnknown(),
				},
			},
		},
	}
}

func (r *TorRelayIdentityRsaResource) Configure(ctx context.Context, req resource.ConfigureRequest, resp *resource.ConfigureResponse) {
}

func (r *TorRelayIdentityRsaResource) Create(ctx context.Context, req resource.CreateRequest, resp *resource.CreateResponse) {
	var data TorRelayIdentityRsaResourceModel

	resp.Diagnostics.Append(req.Plan.Get(ctx, &data)...)
	if resp.Diagnostics.HasError() {
		return
	}

	// Set algorithm
	data.Algorithm = types.StringValue("RSA")

	// Generate 1024-bit RSA private key as required by Tor specification
	// See: https://spec.torproject.org/tor-spec/preliminaries.html#ciphers
	privateKey, err := rsa.GenerateKey(rand.Reader, 1024)
	if err != nil {
		resp.Diagnostics.AddError("RSA Key Generation Error",
			fmt.Sprintf("Unable to generate RSA private key: %s", err))
		return
	}

	// Encode private key as PEM
	privateKeyPem, err := r.encodePrivateKeyPEM(privateKey)
	if err != nil {
		resp.Diagnostics.AddError("Private Key Encoding Error",
			fmt.Sprintf("Unable to encode private key as PEM: %s", err))
		return
	}
	data.PrivateKeyPem = types.StringValue(privateKeyPem)

	// Extract and encode public key
	publicKey := &privateKey.PublicKey
	publicKeyPem, err := r.encodePublicKeyPEM(publicKey)
	if err != nil {
		resp.Diagnostics.AddError("Public Key Encoding Error",
			fmt.Sprintf("Unable to encode public key as PEM: %s", err))
		return
	}
	data.PublicKeyPem = types.StringValue(publicKeyPem)

	// Generate fingerprints
	sha1Fingerprint, sha256Fingerprint, err := r.generateFingerprints(publicKey)
	if err != nil {
		resp.Diagnostics.AddError("Fingerprint Generation Error",
			fmt.Sprintf("Unable to generate public key fingerprints: %s", err))
		return
	}
	data.PublicKeyFingerprintSha1 = types.StringValue(sha1Fingerprint)
	data.PublicKeyFingerprintSha256 = types.StringValue(sha256Fingerprint)

	// Generate ID from SHA1 fingerprint
	data.Id = types.StringValue(fmt.Sprintf("rsa-%s", sha1Fingerprint[:16]))

	tflog.Trace(ctx, "created tor relay identity RSA resource")

	resp.Diagnostics.Append(resp.State.Set(ctx, &data)...)
}

func (r *TorRelayIdentityRsaResource) Read(ctx context.Context, req resource.ReadRequest, resp *resource.ReadResponse) {
	var data TorRelayIdentityRsaResourceModel

	resp.Diagnostics.Append(req.State.Get(ctx, &data)...)
	if resp.Diagnostics.HasError() {
		return
	}

	resp.Diagnostics.Append(resp.State.Set(ctx, &data)...)
}

func (r *TorRelayIdentityRsaResource) Update(ctx context.Context, req resource.UpdateRequest, resp *resource.UpdateResponse) {
	resp.Diagnostics.AddError("Update Not Supported",
		"RSA relay identity keys are immutable. Any changes require resource replacement.")
}

func (r *TorRelayIdentityRsaResource) Delete(ctx context.Context, req resource.DeleteRequest, resp *resource.DeleteResponse) {
	tflog.Trace(ctx, "deleted tor relay identity RSA resource")
}

func (r *TorRelayIdentityRsaResource) encodePrivateKeyPEM(privateKey *rsa.PrivateKey) (string, error) {
	privateKeyBytes := x509.MarshalPKCS1PrivateKey(privateKey)
	privateKeyPem := pem.EncodeToMemory(&pem.Block{
		Type:  "RSA PRIVATE KEY",
		Bytes: privateKeyBytes,
	})
	return string(privateKeyPem), nil
}

func (r *TorRelayIdentityRsaResource) encodePublicKeyPEM(publicKey *rsa.PublicKey) (string, error) {
	publicKeyBytes, err := x509.MarshalPKIXPublicKey(publicKey)
	if err != nil {
		return "", err
	}
	publicKeyPem := pem.EncodeToMemory(&pem.Block{
		Type:  "PUBLIC KEY",
		Bytes: publicKeyBytes,
	})
	return string(publicKeyPem), nil
}

func (r *TorRelayIdentityRsaResource) generateFingerprints(publicKey *rsa.PublicKey) (string, string, error) {
	publicKeyBytes, err := x509.MarshalPKIXPublicKey(publicKey)
	if err != nil {
		return "", "", err
	}

	sha1Sum := sha1.Sum(publicKeyBytes)
	sha256Sum := sha256.Sum256(publicKeyBytes)

	sha1Fingerprint := fmt.Sprintf("%x", sha1Sum)
	sha256Fingerprint := fmt.Sprintf("%x", sha256Sum)

	return sha1Fingerprint, sha256Fingerprint, nil
}
